@@ -1,10 +1,10 @@
 /*
 
-  server_query.c 
+  server_query.c
 
   Author: Pekka Riikonen <priikone@silcnet.org>
 
-  Copyright (C) 2002 Pekka Riikonen
+  Copyright (C) 2002 - 2003 Pekka Riikonen
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
   GNU General Public License for more details.
 
 */
-/* $Id: server_query.c,v 1.19.2.1 2002/12/11 11:49:05 priikone Exp $ */
+/* $Id: server_query.c,v 1.19.2.7 2003/11/06 19:11:47 priikone Exp $ */
 
 #include "serverincludes.h"
 #include "server_internal.h"
@@ -163,7 +163,7 @@ void silc_server_query_send_error(SilcServer server,
 
   /* Send the command reply with error */
   silc_server_send_command_reply(server, query->cmd->sock,
-				 query->querycmd, error, 0, 
+				 query->querycmd, error, 0,
 				 silc_command_get_ident(query->cmd->payload),
 				 argc, data_type, data, data_len);
   va_end(va);
@@ -234,11 +234,15 @@ bool silc_server_query_command(SilcServer server, SilcCommand querycmd,
   switch (querycmd) {
 
   case SILC_COMMAND_WHOIS:
-    /* If we are normal server and query contains nickname, send it
-       directly to router. */
-    if (server->server_type == SILC_SERVER && !server->standalone &&
+    /* If we are normal server and query contains nickname OR query
+       doesn't contain nickname or ids BUT attributes, send it to the
+       router */
+    if (server->server_type != SILC_ROUTER && !server->standalone &&
 	cmd->sock != SILC_PRIMARY_ROUTE(server) &&
-	silc_argument_get_arg_type(cmd->args, 1, NULL)) {
+	 (silc_argument_get_arg_type(cmd->args, 1, NULL) ||
+	 (!silc_argument_get_arg_type(cmd->args, 1, NULL) &&
+	  !silc_argument_get_arg_type(cmd->args, 4, NULL) &&
+	  silc_argument_get_arg_type(cmd->args, 3, NULL)))) {
       silc_server_query_send_router(server, query);
       return TRUE;
     }
@@ -292,7 +296,7 @@ void silc_server_query_send_router(SilcServer server, SilcServerQuery query)
   old_ident = silc_command_get_ident(query->cmd->payload);
   silc_command_set_ident(query->cmd->payload, ++server->cmd_ident);
   tmpbuf = silc_command_payload_encode_payload(query->cmd->payload);
-  silc_server_packet_send(server, 
+  silc_server_packet_send(server,
 			  SILC_PRIMARY_ROUTE(server),
 			  SILC_PACKET_COMMAND, 0,
 			  tmpbuf->data, tmpbuf->len, TRUE);
@@ -329,7 +333,7 @@ void silc_server_query_send_router_reply(void *context, void *reply)
 			   silc_command_get_ident(query->cmd->payload));
     buffer = silc_command_payload_encode_payload(cmdr->payload);
     silc_server_packet_send(server, query->cmd->sock,
-			    SILC_PACKET_COMMAND_REPLY, 0, 
+			    SILC_PACKET_COMMAND_REPLY, 0,
 			    buffer->data, buffer->len, FALSE);
     silc_buffer_free(buffer);
     silc_server_query_free(query);
@@ -357,13 +361,27 @@ void silc_server_query_parse(SilcServer server, SilcServerQuery query)
   switch (query->querycmd) {
 
   case SILC_COMMAND_WHOIS:
+    /* Get requested attributes if set */
+    tmp = silc_argument_get_arg_type(cmd->args, 3, &tmp_len);
+    if (tmp && !query->attrs && tmp_len <= SILC_ATTRIBUTE_MAX_REQUEST_LEN) {
+      query->attrs = silc_attribute_payload_parse(tmp, tmp_len);
+
+      /* When Requested Attributes is present we will assure that this
+	 client cannot execute the WHOIS command too fast.  This would be
+	 same as having SILC_CF_LAG_STRICT. */
+      if (cmd->sock->type == SILC_SOCKET_TYPE_CLIENT &&
+          cmd->sock->user_data)
+	((SilcClientEntry)cmd->sock->user_data)->fast_command = 6;
+    }
+
     /* Get Client IDs if present. Take IDs always instead of nickname. */
     tmp = silc_argument_get_arg_type(cmd->args, 4, &tmp_len);
     if (!tmp) {
 
       /* Get nickname */
       tmp = silc_argument_get_arg_type(cmd->args, 1, &tmp_len);
-      if (!tmp) {
+      if (!tmp && !query->attrs) {
+	/* No nickname, no ids and no attributes - send error */
 	silc_server_query_send_error(server, query,
 				     SILC_STATUS_ERR_NOT_ENOUGH_PARAMS, 0);
 	silc_server_query_free(query);
@@ -371,8 +389,8 @@ void silc_server_query_parse(SilcServer server, SilcServerQuery query)
       }
 
       /* Get the nickname@server string and parse it */
-      if (tmp_len > 128 ||
-	  !silc_parse_userfqdn(tmp, &query->nickname, &query->nick_server)) {
+      if (tmp && ((tmp_len > 128) ||
+	  !silc_parse_userfqdn(tmp, &query->nickname, &query->nick_server))) {
 	silc_server_query_send_error(server, query,
 				     SILC_STATUS_ERR_BAD_NICKNAME, 0);
 	silc_server_query_free(query);
@@ -425,20 +443,7 @@ void silc_server_query_parse(SilcServer server, SilcServerQuery query)
     tmp = silc_argument_get_arg_type(cmd->args, 2, &tmp_len);
     if (tmp && tmp_len == sizeof(SilcUInt32))
       SILC_GET32_MSB(query->reply_count, tmp);
-
-    /* Get requested attributes if set */
-    tmp = silc_argument_get_arg_type(cmd->args, 3, &tmp_len);
-    if (tmp && tmp_len <= SILC_ATTRIBUTE_MAX_REQUEST_LEN) {
-      query->attrs = silc_attribute_payload_parse(tmp, tmp_len);
-
-      /* When Requested Attributes is present we will assure that this
-	 client cannot execute the WHOIS command too fast.  This would be
-	 same as having SILC_CF_LAG_STRICT. */
-      if (cmd->sock->type == SILC_SOCKET_TYPE_CLIENT &&
-	  cmd->sock->user_data)
-	((SilcClientEntry)cmd->sock->user_data)->fast_command = 6;
-    }
-    break;
+   break;
 
   case SILC_COMMAND_WHOWAS:
     /* Get nickname */
@@ -563,6 +568,110 @@ void silc_server_query_parse(SilcServer server, SilcServerQuery query)
   silc_server_query_process(server, query, TRUE);
 }
 
+/* Context for holding clients searched by public key. */
+typedef struct {
+  SilcClientEntry **clients;
+  SilcUInt32 *clients_count;
+} *SilcServerPublicKeyUser, SilcServerPublicKeyUserStruct;
+
+void silc_server_public_key_hash_foreach(void *key, void *context,
+                                         void *user_context)
+{
+  SilcServerPublicKeyUser uc = user_context;
+  SilcClientEntry entry = context;
+
+  /* Nothing was found, just return */
+  if (!context)
+    return;
+
+  (*uc->clients) = silc_realloc((*uc->clients),
+                                sizeof((**uc->clients)) *
+				((*uc->clients_count) + 1));
+  (*uc->clients)[(*uc->clients_count)++] = entry;
+}
+
+/* If clients are set, limit the found clients using the attributes in
+   the query. If clients are not set, try to find some clients using
+   the attributes */
+
+void silc_server_query_check_attributes(SilcServer server,
+                                        SilcServerQuery query,
+                                        SilcClientEntry **clients,
+                                        SilcUInt32 *clients_count) {
+  SilcClientEntry entry;
+  SilcAttributePayload attr;
+  SilcAttribute attribute;
+  SilcAttributeObjPk pk;
+  SilcPublicKey publickey;
+  int i;
+  bool found = FALSE, no_clients = FALSE;
+
+  /* If no clients were found, we only check the attributes
+     if the user wasn't searching for nickname/ids */
+  if (!*clients) {
+    no_clients = TRUE;
+    if (query->nickname || query->ids_count)
+      return;
+  }
+
+  silc_dlist_start(query->attrs);
+  while ((attr = silc_dlist_get(query->attrs)) != SILC_LIST_END) {
+    attribute = silc_attribute_get_attribute(attr);
+    switch (attribute) {
+
+      case SILC_ATTRIBUTE_USER_PUBLIC_KEY:
+	found = TRUE;
+
+	SILC_LOG_DEBUG(("Finding clients by public key attribute"));
+
+	if (!silc_attribute_get_object(attr, &pk, sizeof(pk)))
+	  continue;
+
+	if (!silc_pkcs_public_key_decode(pk.data, pk.data_len,
+	                                 &publickey)) {
+	  silc_free(pk.type);
+	  silc_free(pk.data);
+	  continue;
+	}
+
+	/* If no clients were set on calling this function, we
+	   just search for clients, otherwise we try to limit
+	   the clients */
+	if (no_clients) {
+	  SilcServerPublicKeyUserStruct usercontext;
+
+	  usercontext.clients = clients;
+	  usercontext.clients_count = clients_count;
+
+	  silc_hash_table_find_foreach(server->pk_hash, publickey,
+	                               silc_server_public_key_hash_foreach,
+	                               &usercontext);
+	} else {
+	  for (i = 0; i < *clients_count; i++) {
+	    entry = (*clients)[i];
+
+	    if (!entry->data.public_key)
+	      continue;
+
+	    if (!silc_hash_table_find_by_context(server->pk_hash, publickey,
+		                                 entry, NULL))
+	      (*clients)[i] = NULL;
+	  }
+	}
+	silc_free(pk.type);
+	silc_free(pk.data);
+	silc_pkcs_public_key_free(publickey);
+	break;
+    }
+  }
+
+  if (!found && !query->nickname && !query->ids) {
+    silc_server_query_send_error(server, query,
+                                 SILC_STATUS_ERR_NOT_ENOUGH_PARAMS, 0);
+    silc_server_query_free(query);
+  }
+}
+
 /* Processes the parsed query.  This does the actual finding of the
    queried information and prepares for sending reply to the original
    sender of the query command. */
@@ -591,20 +700,20 @@ void silc_server_query_process(SilcServer server, SilcServerQuery query,
 
   if (query->nickname) {
     /* Get all clients matching nickname from local list */
-    if (!silc_idlist_get_clients_by_hash(server->local_list, 
+    if (!silc_idlist_get_clients_by_hash(server->local_list,
 					 query->nickname, server->md5hash,
 					 &clients, &clients_count))
-      silc_idlist_get_clients_by_nickname(server->local_list, 
+      silc_idlist_get_clients_by_nickname(server->local_list,
 					  query->nickname,
 					  query->nick_server,
 					  &clients, &clients_count);
 
     /* Check global list as well */
     if (check_global) {
-      if (!silc_idlist_get_clients_by_hash(server->global_list, 
+      if (!silc_idlist_get_clients_by_hash(server->global_list,
 					   query->nickname, server->md5hash,
 					   &clients, &clients_count))
-	silc_idlist_get_clients_by_nickname(server->global_list, 
+	silc_idlist_get_clients_by_nickname(server->global_list,
 					    query->nickname,
 					    query->nick_server,
 					    &clients, &clients_count);
@@ -718,6 +827,15 @@ void silc_server_query_process(SilcServer server, SilcServerQuery query,
     }
   }
 
+  /* Check the attributes to narrow down the search by using them. */
+  if (query->attrs)
+    silc_server_query_check_attributes(server, query, &clients,
+                                       &clients_count);
+
+  SILC_LOG_DEBUG(("Querying %d clients", clients_count));
+  SILC_LOG_DEBUG(("Querying %d servers", servers_count));
+  SILC_LOG_DEBUG(("Querying %d channels", channels_count));
+
   /* If nothing was found, then just send the errors */
   if (!clients && !channels && !servers) {
     silc_server_query_send_reply(server, query, NULL, 0, NULL, 0, NULL, 0);
@@ -784,6 +902,11 @@ void silc_server_query_process(SilcServer server, SilcServerQuery query,
 	   client_entry->data.status & SILC_IDLIST_STATUS_NOATTR))
 	continue;
 
+      /* If attributes are present in query, and in the entry and we have
+	 done resolvings already we don't need to resolve anymore */
+      if (query->resolved && query->attrs && client_entry->attrs)
+	continue;
+
       /* Resolve the detailed client information. If client is local we
 	 know that attributes were present and we will resolve directly
 	 from the client. Otherwise resolve from client's owner. */
@@ -965,7 +1088,7 @@ void silc_server_query_resolve(SilcServer server, SilcServerQuery query,
       if (!r) {
 	/* Allocate new temp query list context */
 	query->querylist = silc_realloc(query->querylist,
-					sizeof(*query->querylist) * 
+					sizeof(*query->querylist) *
 					(query->querylist_count + 1));
 	r = &query->querylist[query->querylist_count];
 	query->querylist_count++;
@@ -1132,6 +1255,10 @@ void silc_server_query_send_reply(SilcServer server,
   bool sent_reply = FALSE;
 
   SILC_LOG_DEBUG(("Sending reply to query"));
+  SILC_LOG_DEBUG(("Sending %d clients", clients_count));
+  SILC_LOG_DEBUG(("Sending %d servers", servers_count));
+  SILC_LOG_DEBUG(("Sending %d channels", channels_count));
+  SILC_LOG_DEBUG(("Sending %d errors", query->errors_count));
 
   status = SILC_STATUS_OK;
 
@@ -1143,6 +1270,9 @@ void silc_server_query_send_reply(SilcServer server,
     /* Mark all invalid entries */
     for (i = 0, valid_count = 0; i < clients_count; i++) {
       entry = clients[i];
+      if (!entry)
+	continue;
+
       switch (query->querycmd) {
       case SILC_COMMAND_WHOIS:
 	if (!entry->nickname || !entry->username || !entry->userinfo ||
@@ -1195,7 +1325,7 @@ void silc_server_query_send_reply(SilcServer server,
 
       if (k >= 1)
 	status = SILC_STATUS_LIST_ITEM;
-      if (valid_count > 1 && k == valid_count - 1 
+      if (valid_count > 1 && k == valid_count - 1
 	  && !servers_count && !channels_count && !query->errors_count)
 	status = SILC_STATUS_LIST_END;
       if (query->reply_count && k - 1 == query->reply_count)
@@ -1226,9 +1356,9 @@ void silc_server_query_send_reply(SilcServer server,
 		       server->server_name, len);
 	}
       }
-      
+
       switch (query->querycmd) {
-	
+
       case SILC_COMMAND_WHOIS:
 	{
 	  unsigned char idle[4], mode[4];
@@ -1248,11 +1378,11 @@ void silc_server_query_send_reply(SilcServer server,
 
 	  if (cmd->sock->type == SILC_SOCKET_TYPE_CLIENT)
 	    channels =
-	      silc_server_get_client_channel_list(server, entry, FALSE, 
+	      silc_server_get_client_channel_list(server, entry, FALSE,
 						  FALSE, &umode_list);
 	  else
 	    channels =
-	      silc_server_get_client_channel_list(server, entry, TRUE, 
+	      silc_server_get_client_channel_list(server, entry, TRUE,
 						  TRUE, &umode_list);
 
 	  if (memcmp(entry->data.fingerprint, fempty, sizeof(fempty)))
@@ -1270,8 +1400,8 @@ void silc_server_query_send_reply(SilcServer server,
 	  if (query->attrs) {
 	    if (!entry->attrs && SILC_IS_LOCAL(entry)) {
 	      tmpattrs = silc_server_query_reply_attrs(server, query, entry);
-	      entry->attrs = silc_memdup(tmpattrs->data, tmpattrs->len);
-	      entry->attrs_len = tmpattrs->len;
+	      entry->attrs = silc_buffer_steal(tmpattrs, &len);
+	      entry->attrs_len = len;
 	      silc_buffer_free(tmpattrs);
 	    }
 	    attrs = entry->attrs;
@@ -1284,7 +1414,7 @@ void silc_server_query_send_reply(SilcServer server,
 					 2, idp->data, idp->len,
 					 3, nh, strlen(nh),
 					 4, uh, strlen(uh),
-					 5, entry->userinfo, 
+					 5, entry->userinfo,
 					 strlen(entry->userinfo),
 					 6, channels ? channels->data : NULL,
 					 channels ? channels->len : 0,
@@ -1353,8 +1483,8 @@ void silc_server_query_send_reply(SilcServer server,
 				       2, idp->data, idp->len,
 				       3, nh, strlen(nh),
 				       4, uh, strlen(uh),
-				       5, entry->userinfo, 
-				       entry->userinfo ? 
+				       5, entry->userinfo,
+				       entry->userinfo ?
 				       strlen(entry->userinfo) : 0);
 	sent_reply = TRUE;
 	break;
@@ -1371,7 +1501,7 @@ void silc_server_query_send_reply(SilcServer server,
       /* Not one valid entry was found, send error.  If nickname was used
 	 in query send error based on that, otherwise the query->errors
 	 already includes proper errors. */
-      if (query->nickname)
+      if (query->nickname || (!query->nickname && !query->ids && query->attrs))
 	silc_server_query_add_error(server, query, TRUE, 1,
 				    SILC_STATUS_ERR_NO_SUCH_NICK);
     }
@@ -1387,15 +1517,18 @@ void silc_server_query_send_reply(SilcServer server,
     k = 0;
     for (i = 0; i < servers_count; i++) {
       entry = servers[i];
-      
+
       if (k >= 1)
 	status = SILC_STATUS_LIST_ITEM;
+      if (servers_count == 1 && status != SILC_STATUS_OK && !channels_count &&
+	  !query->errors_count)
+	status = SILC_STATUS_LIST_END;
       if (servers_count > 1 && k == servers_count - 1 && !channels_count &&
 	  !query->errors_count)
 	status = SILC_STATUS_LIST_END;
       if (query->reply_count && k - 1 == query->reply_count)
 	status = SILC_STATUS_LIST_END;
-      
+
       SILC_LOG_DEBUG(("%s: server %s",
 		      (status == SILC_STATUS_OK ?         "   OK" :
 		       status == SILC_STATUS_LIST_START ? "START" :
@@ -1408,13 +1541,13 @@ void silc_server_query_send_reply(SilcServer server,
       idp = silc_id_payload_encode(entry->id, SILC_ID_SERVER);
       silc_server_send_command_reply(server, cmd->sock, SILC_COMMAND_IDENTIFY,
 				     status, 0, ident, 2,
-				     2, idp->data, idp->len, 
-				     3, entry->server_name, 
-				     entry->server_name ? 
+				     2, idp->data, idp->len,
+				     3, entry->server_name,
+				     entry->server_name ?
 				     strlen(entry->server_name) : 0);
       silc_buffer_free(idp);
       sent_reply = TRUE;
-      
+
       if (status == SILC_STATUS_LIST_END)
 	break;
       k++;
@@ -1431,15 +1564,18 @@ void silc_server_query_send_reply(SilcServer server,
     k = 0;
     for (i = 0; i < channels_count; i++) {
       entry = channels[i];
-      
+
       if (k >= 1)
 	status = SILC_STATUS_LIST_ITEM;
+      if (channels_count == 1 && status != SILC_STATUS_OK &&
+	  !query->errors_count)
+	status = SILC_STATUS_LIST_END;
       if (channels_count > 1 && k == channels_count - 1 &&
 	  !query->errors_count)
 	status = SILC_STATUS_LIST_END;
       if (query->reply_count && k - 1 == query->reply_count)
 	status = SILC_STATUS_LIST_END;
-      
+
       SILC_LOG_DEBUG(("%s: channel %s",
 		      (status == SILC_STATUS_OK ?         "   OK" :
 		       status == SILC_STATUS_LIST_START ? "START" :
@@ -1452,13 +1588,13 @@ void silc_server_query_send_reply(SilcServer server,
       idp = silc_id_payload_encode(entry->id, SILC_ID_CHANNEL);
       silc_server_send_command_reply(server, cmd->sock, SILC_COMMAND_IDENTIFY,
 				     status, 0, ident, 2,
-				     2, idp->data, idp->len, 
-				     3, entry->channel_name, 
-				     entry->channel_name ? 
+				     2, idp->data, idp->len,
+				     3, entry->channel_name,
+				     entry->channel_name ?
 				     strlen(entry->channel_name) : 0);
       silc_buffer_free(idp);
       sent_reply = TRUE;
-      
+
       if (status == SILC_STATUS_LIST_END)
 	break;
       k++;
@@ -1502,6 +1638,8 @@ void silc_server_query_send_reply(SilcServer server,
 
       if (k >= 1)
 	status = SILC_STATUS_LIST_ITEM;
+      if (query->errors_count == 1 && status != SILC_STATUS_OK)
+	status = SILC_STATUS_LIST_END;
       if (query->errors_count > 1 && k == query->errors_count - 1)
 	status = SILC_STATUS_LIST_END;
       if (query->reply_count && k - 1 == query->reply_count)
@@ -1512,7 +1650,7 @@ void silc_server_query_send_reply(SilcServer server,
 		       status == SILC_STATUS_LIST_START ? "START" :
 		       status == SILC_STATUS_LIST_ITEM  ? " ITEM" :
 		       status == SILC_STATUS_LIST_END  ?  "  END" :
-		       "      : "), 
+		       "      : "),
 		      silc_get_status_message(query->errors[i].error),
 		      query->errors[i].error));
 
@@ -1596,7 +1734,7 @@ SilcBuffer silc_server_query_reply_attrs(SilcServer server,
     case SILC_ATTRIBUTE_STATUS_FREETEXT:
       /* Put STATUS_FREETEXT.  We just tell in the message that we are
 	 replying on behalf of the client. */
-      tmp = 
+      tmp =
 	"This information was provided by the server on behalf of the user";
       buffer = silc_attribute_payload_encode(buffer, attribute,
 					     SILC_ATTRIBUTE_FLAG_VALID,
@@ -1645,7 +1783,7 @@ SilcBuffer silc_server_query_reply_attrs(SilcServer server,
       if (attribute == SILC_ATTRIBUTE_SERVER_PUBLIC_KEY ||
 	  attribute == SILC_ATTRIBUTE_SERVER_DIGITAL_SIGNATURE)
 	break;
-      
+
       /* For other attributes we cannot reply so mark it invalid */
       buffer = silc_attribute_payload_encode(buffer, attribute,
 					     SILC_ATTRIBUTE_FLAG_INVALID,
